# Generated by Django 3.0 on 2022-11-25 16:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("rxn", "0002_reaction_latex"),
    ]

    operations = [
        migrations.AddField(
            model_name="reaction",
            name="ordered_text",
            field=models.CharField(default="a", editable=False, max_length=256),
            preserve_default=False,
        ),
    ]
